﻿/**********************************************************/
/*  Copyright © YoAmbulante (alex.nino@yoambulante.com)   */
/*                                                        */
/*  MidiFilesReader is a class for reading MIDI content   */
/*  from a .mid file which could be used for              */
/*  synchronizing animations and dynamic sound.           */
/*                                                        */
/*  Visit http://www.yoambulante.com to clarify any       */
/*  doubt with this code.                                 */
/*  Feel free to use this code comercially. You may       */
/*  modify this class and adjust it to your requirements. */
/*  If you make any improvement, or have found any bug    */
/*  and fixed it please shared it with the community,     */
/*  send me an email.                                     */
/*                                                        */
/*  MidiFilesReader v1.0 - Alex Nino (Nov 30th, 2008)     */
/*  MidiFilesReader v1.1 - Alex Nino (May 15th, 2009)     */
/*    - bug detected with Running status with different   */
/*      notes on NOTE_ON Midi event (fixed)               */
/*    - .loader property available externally             */
/*  MidiFilesReader v1.2 - Alex Nino (Oct 15th, 2010)     */
/*    - when reading a midi event, if it is NOTE_ON with  */
/*      velocity zero, then it is converted to NOTE_OFF   */
/*    - improvements made for Flash Player 10.            */
/*                                                        */
/**********************************************************/

package com.yoambulante.midifiles {
	import flash.events.EventDispatcher;
	import flash.events.HTTPStatusEvent;
	import flash.events.IEventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.ProgressEvent;
	import flash.events.SecurityErrorEvent;
	import flash.net.URLLoaderDataFormat;
	
	import flash.net.URLLoader;
	import flash.net.URLRequest;
	import flash.utils.ByteArray;
	import flash.events.Event;
	import flash.utils.Endian;
	
	/**
	* Dispatched as soon as the midi header file was found when reading the binary content.
	*
	* @eventType com.yoambulante.midifiles.MidiFilesReader.ON_HEADER_LOADED
	*/
	[Event(name = "ON_HEADER_LOADED", type = "flash.events.Event")]
	
	/**
	* Dispatched when the midi file was read successfully and content valid data.
	*
	* @eventType com.yoambulante.midifiles.MidiFilesReader.ON_DATA_LOADED
	*/
	[Event(name = "ON_DATA_LOADED", type = "flash.events.Event")]
	
	/**
	* Dispatched if during the reading process there was any unexpected byte or error.
	*
	* @eventType com.yoambulante.midifiles.MidiFilesReader.ON_DATA_ERROR
	*/
	[Event(name="ON_DATA_ERROR", type="flash.events.Event")]
	
	/**
	 * This class reads a standard MIDI file (.mid) and exposes all events through all different tracks found in the midi file
	 * 
	 * @author Alex Nino
	 */
	public class MidiFilesReader extends EventDispatcher {
		
		/**
		*
		* @eventType ON_HEADER_LOADED
		*/
		static public const ON_HEADER_LOADED:String = "ON_HEADER_LOADED";
		/**
		*
		* @eventType ON_DATA_LOADED
		*/
		static public const ON_DATA_LOADED:String = "ON_DATA_LOADED";
		/**
		*
		* @eventType ON_DATA_LOADED
		*/
		static public const ON_DATA_ERROR:String = "ON_DATA_ERROR";		
		
		static private const STATUS_EMPTY:uint = 0;
		static private const STATUS_MTHD_FOUND:uint = 1;		
		static private const STATUS_HEADER_LOADED:uint = 2;
		static private const STATUS_FILE_LOADED:uint = 3;
		
		private var status:uint;
		
		private var data:ByteArray;		
		private var _loader:URLLoader;
		private var trackIndex:int;		
		private var readtrack:int;	
		private var _downloading:Boolean;
		
		private var _trackNumber:int;	
		private var _resolution:uint;	
		private var _subResolution:uint;	
		private var _tracksType:int;	
		
		private var lastEvent:MidiMessage;
		private var currentTrack:MidiTrack;
		
		private var _tracks:Vector.<MidiTrack>;
		private var _totalTracks:uint;		
		
		function MidiFilesReader() {
			_loader = new URLLoader(); //used for loading externally a .mid file.
			_loader.dataFormat = URLLoaderDataFormat.BINARY;	//ensure that the loaded content is binary and not text.	
			_loader.addEventListener(Event.COMPLETE, completeHandler);
			_loader.addEventListener(IOErrorEvent.IO_ERROR, ioErrorHandler);
			reset();
		}
		
		/**
		 * reset all internal data previously loaded
		 */
		private function reset():void {
			if (_downloading) {
				loader.close();
			}
			data = new ByteArray();
			_downloading = false;
			_totalTracks = 0;
			_tracks = new Vector.<MidiTrack>();
			readtrack = -1;
			_resolution = 0;
			trackIndex = -1;
			_tracksType = 0;
			_subResolution = 0;
			_trackNumber = 0;			
			status = STATUS_EMPTY;
		}
		
		/**
		 * Loads externally a binary file with a MIDI standard estructure
		 * @param	filename url of the .mid file, it can be both a local file or a remote location from a web server (see security limitations)
		 */
		public function loadFile(filename:String):void {			
			reset(); 			
			_downloading = true;
			_loader.load(new URLRequest(filename));						
		}

		
		private function completeHandler(event:Event):void {		
			_downloading = false;
			loadFromBytes(_loader.data);			
		}
		
		private function ioErrorHandler(event:IOErrorEvent):void {
			//trace("ioErrorHandler: " + event);
			_downloading = false;			
			dispatchEvent(event);
			dispatchEvent(new Event(ON_DATA_ERROR));
		}
		
		/**
		 * MIDI file track type, 0: single track file format. 1:multiple track file format. 2:multiple song file format.
		 * @return String telling a detailed description
		 */
		public function getTrackTypeDescription():String {			
			switch (_tracksType) {
				case 0 :
					return "single track file format.";
					break;
				case 1 :
					return "multiple track file format.";
					break;
				case 2 :
					return "multiple song file format.";
					break;
			}
			return "unknown";
		}
		
		/**
		 * Load a MIDI file structure from a <code>ByteArray</code>
		 * @param	bin_file <code>ByteArray</code> file, note that you must specify the starting position of the file, this function doesn't set bin_file.position = 0
		 */
		public function loadFromBytes(bin_file:ByteArray):void {
			reset();
			data = bin_file;
			data.endian = Endian.BIG_ENDIAN; //midi files data values are always stored in Big-Endian 			
			//trace("MIDIFILE size:" + data.length + " bytes, reading...");			
			readHeaderChunk();
			if (status == STATUS_HEADER_LOADED){
				dispatchEvent(new Event(ON_HEADER_LOADED));
			} else {
				dispatchEvent(new Event(ON_DATA_ERROR));
			}
		}
		
		/**
		 * You control the workflow and when the events are read. As long as you're ready for carry on reading the file just call <code>nextEvent</code>
		 * @return returns <code>true</code> if any data was read, otherwise <code>false</code>, normally when it reaches the end of the file.
		 */
		public function nextEvent():Boolean{
			if (status == STATUS_HEADER_LOADED) { //header loaded.
				if (data.position < data.length){
					if ((readtrack == -1 || data.position >= readtrack) && readNextChunk()) {												
						trackIndex++; //track header was read
					}
					readNextEvent();
					return true;					
				} else {
					//trace("end of file.");
					status = STATUS_FILE_LOADED;
					dispatchEvent(new Event(ON_DATA_LOADED));
				}
			} else {
				//trace("invalid midi file.")				
				dispatchEvent(new Event(ON_DATA_ERROR));
			}
			return false;
		}		
		
		private function readHeaderChunk():Boolean {
			data.position = 0; //reset position to the begining of the file
			if (data.readUnsignedInt() == 0x4D546864) { //MThd -> 0x4D546864
				//trace("HeaderChunk(MThd) ok.");
				status = STATUS_MTHD_FOUND;
			} else {
				//trace("Invalid MIDI file format.");
				return false;
			}			
			if (data.readUnsignedInt() == 6 && data.position < data.length-7) { //read header length, must be always 6 bytes.				
				//trace("header length ok.");								
				_tracksType = data.readUnsignedShort();
				_trackNumber = data.readUnsignedShort();
				//trace("number of track : " + _trackNumber);
				_resolution = data.readUnsignedShort();
				if (_resolution & 0x8000) {
					//frames per second
					_subResolution = _resolution & 0x00FF; //read only the 0x7FFF part of resolution
					_resolution = _resolution & 0x7F00; //read only the 0x7FFF part of resolution
					/*trace("frames per second : " + _resolution);																	
					trace("ticks per frame : " + _subResolution);*/
				} else {
					//it is in ticks per beat or pulses per quarter note
					//trace("ticks per beat : " + _resolution); 
				}
				status = STATUS_HEADER_LOADED;
				return true;
			}
			return false;
		}		
		
		private function getVarLength():uint {
			var vlen:uint = 0;				
			var b:uint; //Byte
			do {
				b = data.readUnsignedByte();
				vlen = (vlen << 7) | (b & 0x7F);
			} while (b & 0x80);	
			return vlen;
		}
		
		private function readNextEvent():void {
			var b:uint; //Byte			
			var t:uint; //type			
			var bytes:ByteArray;
			var delta_time:uint = getVarLength();							
			b = data.readUnsignedByte();			
			var e:MidiMessage;			
			if (b >> 4 == 0xF) { //it is a System Exclusive Message (MetaEvent).
				if (b == 0xFF) { //meta event					
					t = data.readUnsignedByte();					
					switch (t) {
						case 0:	//Sequence Number							
							if (data.readUnsignedByte() == 2) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, data.readUnsignedByte(), data.readUnsignedByte());
							}							
							break;				
						case 1 : //Text Event							
						case 2 : //Copyright Notice
						case 3 : //Sequence/Track Name
						case 4 : //Instrument Name
						case 5 : //Lyrics
						case 6 : //Marker
						case 7 : //Cue Point
						case 127 : //Sequencer Specific as text value							
							e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, 0, 0, data.readMultiByte(getVarLength(), "iso-8859-1")); //it is a text metatag
							break;		
						case 32 : //MIDI Channel Prefix
							if (data.readUnsignedByte() == 1) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, data.readUnsignedByte(), 0, 0);
							}							
							break;	
						case 47 : //End Of Track
							if (data.readUnsignedByte() == 0) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, 0, 0);
								readtrack = -1; //read next track
							}							
							break;
						case 81 : //Set Tempo
							if (data.readUnsignedByte() == 3) {
								var bmp:uint = data.readUnsignedByte();								
								bmp = (bmp << 8) | (data.readUnsignedByte() & 0xFF);
								bmp = (bmp << 8) | (data.readUnsignedByte() & 0xFF);
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, bmp, 0);
							}						
							break;
						case 84 : //SMPTE Offset
							if (data.readUnsignedByte() == 5) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, data.readUnsignedByte(), data.readUnsignedByte(), "", data.readUnsignedByte(), data.readUnsignedByte(), data.readUnsignedByte());
							}							
							break;		
						case 88 : //Time Signature							
							if (data.readUnsignedByte() == 4) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, data.readUnsignedByte(), data.readUnsignedByte(), "", data.readUnsignedByte(), data.readUnsignedByte());
							}							
							break;	
						case 89 : //Key Signature (Key and Scale)							
							if (data.readUnsignedByte() == 2) {
								e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, data.readUnsignedByte(), data.readUnsignedByte());
							}
							break;												
						default : //undefined metaevent														
							bytes= new ByteArray(); //review this code, it may not apply for all cases.
							bytes.writeBytes(data, 0, data.readUnsignedByte());
							e = new MidiMessage(MidiMessage.CATEGORY_METAEVENT, delta_time, t, 0, 0, 0, "", 0, 0, 0, bytes);
							break;
					}
				} else if (b == 0xF7 || b == 0xF0) { //System Exclusive Events
					bytes= new ByteArray();
					bytes.writeBytes(data, 0, getVarLength());
					e = new MidiMessage(MidiMessage.CATEGORY_SYSTEM_EXCLUSIVE, delta_time, t, 0, 0, 0, "", 0, 0, 0, bytes);	
				}
			} else {
				t = b >> 4;
				if (t >= 0x8 && t <= 0xE){ //create a MIDI event
					e = new MidiMessage(MidiMessage.CATEGORY_MIDI_CONTROL, delta_time, t, b & 0xF, data.readUnsignedByte(), (t != 12 && t != 13)?data.readUnsignedByte():0);
					lastEvent = e; //save previous event.
				} else { //is it a running event?
					e = new MidiMessage(lastEvent.category, delta_time, lastEvent.type, lastEvent.channel, b, lastEvent.category!=1||(lastEvent.type != 12 && lastEvent.type != 13)?data.readUnsignedByte():0);
				}
			}
			if (e) {				
				//trace(">>>> event [TRACK:"+track_index+"], category:" + e.category + ", type:" + e.type + ", DELTA_TIME:"+e.delta_time+", name:"+MidiMessage.getTypeDescription(e));						
				currentTrack.addEvent(e);
			}
		}
		private function readNextChunk():Boolean {
			if (data.readUnsignedInt() != 0x4D54726B) { //MTrk => 0x4D54726B				
				//trace("Invalid Chunk block format.");
				return false;
			}						
			readtrack = data.position + data.readUnsignedInt() + 4; //plus 4 is the end track
			currentTrack = new MidiTrack();
			_tracks.push(currentTrack);
			_totalTracks++;
			return true;
		}

		/**
		 * returns total amount of available tracks after loading.
		 */
		public function get totalTracks():uint { return _totalTracks; }
		
		/**
		 * resolution normally means how many ticks per quater note
		 */
		public function get resolution():uint { return _resolution; }
		
		/**
		 * subResolution used when the midi file time is in frames per second
		 */
		public function get subResolution():uint { return _subResolution; }
		
		/**
		 * MIDI file track type, 0: single track file format. 1:multiple track file format. 2:multiple song file format.		 
		 */
		public function get tracksType():int { return _tracksType; }
		
		/**
		 * This is the <code>URLLoader</code> instance used when loading external files. Use this property to attach your own event listeners and have more control about the loading process workflow.
		 */
		public function get loader():URLLoader { return _loader; }
		
		/**
		 * gets a track from list
		 * @param	index must be a uint &gt;= 0 and &lt; totalTracks
		 * @return return null if index is equal or higher than totalTracks
		 */
		public function getTrack(index:uint):MidiTrack {
			if (index < _totalTracks) {
				return _tracks[index];
			}
			return null;
		}
	}
}