package com.yoambulante.midifiles {
	import flash.utils.ByteArray;
	/**
	 * MidiMessage contains information about the midi event (message) on an especific midi track
	 * @author Alex Nino (yoambulante.com)
	 */
	public class MidiMessage {
		static public const CATEGORY_MIDI_CONTROL:uint = 1;
		static public const CATEGORY_SYSTEM_EXCLUSIVE:uint = 2;
		static public const CATEGORY_METAEVENT:uint = 3;
		/**
		 * extracts the event type description
		 * @param	e a valid MidiType event (message)
		 * @return return a string with detailed description of the message type
		 */
		static public function getTypeDescription(e:MidiMessage):String {
			if (e.category == CATEGORY_MIDI_CONTROL){
				switch (e.type) {
					case 0x8: return "Note Off, param1: note = "+e.param1+", param2: velocity = "+e.param2;
					case 0x9: return "Note On, param1: note = "+e.param1+", param2: velocity = "+e.param2;
					case 0xA: return "Note Aftertouch, param1: note = "+e.param1+", param2: aftertouch value = "+e.param2;
					case 0xB: return "Controller, ("+getControllerDescription(e)+") - param1: "+e.param1+", param2: "+e.param2;
					case 0xC: return "Program Change, param1: program number "+e.param1+", param2: not used";
					case 0xD: return "Channel Aftertouch, param1: aftertouch value "+e.param1+", param2: not used";
					case 0xE: return "Pitch Bend, param1: pitch value (LSB) "+e.param1+", param2: pitch value (MSB) "+e.param2;					
					default : return "unknown midi controller event, param1:"+e.param1+", param2:"+e.param2;
				}
			} else if (e.category == CATEGORY_METAEVENT) {
				switch (e.type) {
					case 0:	return "Sequence Number";
					case 1 : return "Text MetaTag ("+e.str_data+")";
					case 2 : return "Copyright Notice ("+e.str_data+")";
					case 3 : return "Sequence/Track Name ("+e.str_data+")";
					case 4 : return "Instrument Name ("+e.str_data+")";
					case 5 : return "Lyrics ("+e.str_data+")";
					case 6 : return "Marker ("+e.str_data+")";
					case 7 : return "Cue Point ("+e.str_data+")";
					case 32 : return "MIDI Channel Prefix";
					case 47 : return "End Of Track";
					case 81 : return "Set Tempo - Microseconds per quarter note ("+e.param1+")";
					case 84 : return "SMPTE Offset p1("+e.param1+"), p2("+e.param2+"), p3("+e.param3+"), p4("+e.param4+"), p5("+e.param5+")";
					case 88 : return "Time Signature - Numerator("+e.param1+"), Denominator("+e.param2+"), Metronome-pulse("+e.param3+"), 32nds("+e.param4+")";
					case 89 : return "Key Signature";
					case 127 : return "Sequencer Specific as text value ("+e.str_data+")";
					default : return "unknown meta event param1:"+e.param1+", param2:"+e.param2;
				}
			}  else if (e.category == CATEGORY_SYSTEM_EXCLUSIVE) {
				return "SYSTEM EXCLUSIVE EVENT bin_block "+e.bin_data.length+" bytes";
			}
			return "unknown or non-common";
		}
		
		/**
		 * Extracts the controller event description based on param1
		 * @param	e a valid MidiType event
		 * @return return a string with detailed description of the controller event
		 */
		static public function getControllerDescription(e:MidiMessage):String {
			if (e.type == 0xB){
				switch (e.param1) {
					case 0 : return "Bank Select";
					case 1 : return "Modulation";
					case 2 : return "Breath Controller";
					case 4 : return "Foot Controller";
					case 5 : return "Portamento Time";
					case 6 : return "Data Entry (MSB)";
					case 7 : return "Main Volume";
					case 8 : return "Balance";
					case 10 : return "Pan";
					case 11 : return "Expression Controller";
					case 12 : return "Effect Control 1";
					case 13 : return "Effect Control 2";
					case 16 :
					case 19 : return "General-Purpose Controllers 1-4";
					case 32 :
					case 63 : return "LSB for controllers 0-31";
					case 64 : return "Damper pedal (sustain)";
					case 65 : return "Portamento";
					case 66 : return "Sostenuto";
					case 67 : return "Soft Pedal";
					case 68 : return "Legato Footswitch";
					case 69 : return "Hold 2";
					case 70 : return "Sound Controller 1 (default: Timber Variation)";
					case 71 : return "Sound Controller 2 (default: Timber/Harmonic Content)";
					case 72 : return "Sound Controller 3 (default: Release Time)";
					case 73 : return "Sound Controller 4 (default: Attack Time)";
					case 74 :
					case 79 : return "Sound Controller 6-10";
					case 80 :
					case 83 : return "General-Purpose Controllers 5-8";
					case 84 : return "Portamento Control";
					case 91 : return "Effects 1 Depth (formerly External Effects Depth)";
					case 92 : return "Effects 2 Depth (formerly Tremolo Depth)";
					case 93 : return "Effects 3 Depth (formerly Chorus Depth)";
					case 94 : return "Effects 4 Depth (formerly Celeste Detune)";
					case 95 : return "Effects 5 Depth (formerly Phaser Depth)";
					case 96 : return "Data Increment";
					case 97 : return "Data Decrement";
					case 98 : return "Non-Registered Parameter Number (LSB)";
					case 99 : return "Non-Registered Parameter Number (MSB)";
					case 100 : return "Registered Parameter Number (LSB)";
					case 101 : return "Registered Parameter Number (MSB)";
					case 121 :
					case 127 : return "Mode Messages";
				}
			} else {
				return "e:MidiMessage type is not a Controller";
			}
			return "unknown";
		}		
		
		/**
		 * returns a Number which represents the tempo BPM beats per minute.
		 * @param	tempo Set-Tempo midi event message
		 * @param	signature Time-Signature midi event message
		 * @return
		 */
		static public function getTempoBPM(tempo:MidiMessage, signature:MidiMessage):Number {
			if (tempo.category == CATEGORY_METAEVENT && tempo.type == 81 && signature.category == CATEGORY_METAEVENT && signature.type == 88) {
				const MICROSECONDS_PER_MINUTE:uint = 60000000;
				return (MICROSECONDS_PER_MINUTE / tempo.param1) * (signature.param1 / 4.);
			}
			trace("Invalid metaevent (Set Tempo & Time Siganture)");
			return 0;			
		}
		/**
		 * returns tempo of a metaevent (set tempo) in beats per minute
		 * @param	e a valid MidiType event
		 * @return return a Number which represents the tempo of a metaevent (set tempo) in beats per minute
		 */
		static public function _getTempoBPM(e:MidiMessage):Number { //when it is 4/4
			if (e.category == CATEGORY_METAEVENT && e.type == 81) {
				const MICROSECONDS_PER_MINUTE:uint = 60000000;
				return MICROSECONDS_PER_MINUTE / e.param1;
			}
			trace("e:MidiMessage is not a metaevent (Set Tempo)");
			return 0;			
		}
		static public function getTempoMPQN(e:MidiMessage):Number {
			if (e.category == CATEGORY_METAEVENT && e.type == 81) {
				const MICROSECONDS_PER_MINUTE:uint = 60000000;
				return MICROSECONDS_PER_MINUTE / (MICROSECONDS_PER_MINUTE / e.param1);
			}
			trace("e:MidiMessage is not a metaevent (Set Tempo)");
			return 0;
		}
		
		
		private var _delta_time:uint;
		private var _type:uint;
		private var _category:uint;
		private var _channel:uint;
		private var _param1:uint;
		private var _param2:uint;
		private var _param3:uint;
		private var _param4:uint;
		private var _param5:uint;
		private var _str_data:String;
		private var _bin_data:ByteArray;
		private var _processed:Boolean;
		
		public function MidiMessage(cate:uint, dtime:uint, type:uint, channel:uint, param1:uint, param2:uint, str:String = "", param3:uint = 0, param4:uint = 0, param5:uint = 0, bin:ByteArray = null) {
			_delta_time = dtime;
			_category = cate;
			_type = type;
			_channel = channel;
			_param1 = param1;
			_param2 = param2;
			if (_category == CATEGORY_MIDI_CONTROL && _type == 0x9 && _param2 == 0) { //turn this NOTE ON to OFF (NOTE ON with velocity zero is equivalent to NOTE OFF).
				_type = 0x8;
			}
			_param3 = param3;
			_param4 = param4;
			_param5 = param5;
			_str_data = str;
			_bin_data = bin;
			_processed = false;
		}
		
		public function clone():MidiMessage {
			var b:ByteArray;
			if (_bin_data) {
				b = new ByteArray();
				var pos:uint = _bin_data.position;
				_bin_data.position = 0;
				b.writeBytes(_bin_data);
				_bin_data.position = pos;
			}
			return new MidiMessage(_category, _delta_time, _type, _channel, _param1, _param2, _str_data, _param3, _param4, _param5, b);
		}
		
		/**
		 * Use this function to extract a full detailed descript about the event.
		 * @return full detailed information about the event
		 */
		public function toString():String {
			return "MidiMessage [category:"+this._category+"(0x"+this._category.toString(16).toUpperCase()+") ch:"+this._channel+", type:"+this._type+"(0x"+this._type.toString(16).toUpperCase()+")] delta_time["+_delta_time+"] => " + getTypeDescription(this);
		}
		
		/**
		 * time ralative with previous event, note zero means must be played at the same time of previous event, when you read this property it turns the .processed to true
		 */
		public function get delta_time():uint { _processed = true; return _delta_time; }
		
		/**
		 * Allows you to override (set a new value) the delta time value of the event.
		 * @param	val new delta time value
		 */
		public function resetDeltaTime(val:uint):void { _delta_time = val; }
		
		/**
		 * Event type, use function MidiMessage.getTypeDescription to know more about a especific event.
		 */
		public function get type():uint { return _type; }
		
		/**
		 * Midi event channel
		 */
		public function get channel():uint { return _channel; }
		
		/**
		 * depending on what kind of event this value must be used differently, normally a value between 0 and 127
		 */
		public function get param1():uint { return _param1; }
		
		/**
		 * depending on what kind of event this value must be used differently, normally a value between 0 and 127
		 */
		public function get param2():uint { return _param2; }
		
		/**
		 * param3 is used by the MetaEvent SMPTE Offset as Seconds (param1 as Hours and param2 as Minutes) and Time Signature MetaEvent param3 is Metronome-pulse (param1 Numerator and param2 Denominator)
		 */
		public function get param3():uint { return _param3; }
		
		/**
		 * param4 is used by the MetaEvent SMPTE Offset as Frames (param1 as Hours and param2 as Minutes) and Time Signature MetaEvent param4 is 32nds (param1 Numerator and param2 Denominator)
		 */
		public function get param4():uint { return _param4; }
		
		/**
		 * param5 is only used by the MetaEvent SMPTE Offset as SubFrames (param1 as Hours and param2 as Minutes)
		 */
		public function get param5():uint { return _param5; }
		
		/**
		 * this is the event category which could be only one of three CATEGORY_MIDI_CONTROL, CATEGORY_SYSTEM_EXCLUSIVE, CATEGORY_METAEVENT.
		 */
		public function get category():uint { return _category; }		
		
		/**
		 * String data used in expecific MetaEvent
		 */
		public function get str_data():String { return _str_data; }
		
		/**
		 * Binary data used in System Exclusive Events and in unknown metaevents
		 */
		public function get bin_data():ByteArray { return _bin_data; }
		
		/**
		 * indicates whenever this event message was read
		 */
		public function get processed():Boolean { return _processed; }
		
	}

}