package com.yoambulante.midifiles {
	import flash.display.GradientType;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.SampleDataEvent;
	import flash.geom.Point;
	import flash.media.Sound;
	import flash.media.SoundChannel;
	import flash.utils.ByteArray;
	import flash.utils.Endian;
	
	/**
	 * This class uses a waveform (samples) and change its pitch using hermite interpolation for generating different musical notes tones.
	 * 
	 * @author Alex Nino
	 */
	public class SamplerKey {						
		private var noteSamples:Vector.<Point>; //notes already generated.
		private var _note:int; //this is the key index (where in the keyboard this note is), waveform samples must start in A440.
		private var _noteLP:Number; //this value is only used when applying LOW-PASS filter.
		private var _midiNote:int; //this is the global midi position of this key-note.
		private var _noteFactor:Number; //how much from A4 this note needs to be stretched or shrunk (used when interpolating samples)		
		private var newLength:uint; //new length after stretch based on pair samples (both left and right samples counts as one)		
		private var startFading:uint; //a limit for fade out (decay)	
		private var fadeOut:Number; //use to release the note smoothly	(decay)	
		private var fadeIn:Number; //use to play the key smoothly	(decay)	
		private var velFact:Number;
		private var fades:Number;
		private var _playing:Boolean;		
		
		//internal variables used when interpolating waveform
		private var extractor:SamplesExtractor;
		private var totalSamples:uint; //it must be extractor.totalSamples, store this value locally to avoid calling the (get) function from extractor instance all the time.
		private var pos:Number;
		private var i:uint;
		private var step:Number;
		//sample values calculated on hermite same interpolation.
		private var s1:Point;
		private var s2:Point;
		private var s3:Point;
		private var s4:Point;		
		private var hermite:Number; //this is the hermite value interpolated
		private var released:Boolean; //is this key-note virtually pressed?
		private var _lastSample:Point; //last generated sample (with no volume changes applied)
		private var _output:Point; //last generated sample with fades.
		private var sindex:uint;
		private var _velocity:uint;
		
		/**
		 * 
		 * @param	ext SamplesExtractor which has already extracted all samples from waveform (Sound file based on a waveform A440 frequency)
		 * @param	key Key-note representation in the Keyboard
		 * @param	midi_key Key-note representation in the Keyboard related to MIDI protocol
		 */ 
		public function SamplerKey(ext:SamplesExtractor, key:int, midi_key:int) {
			_playing = false;
			extractor = ext;			
			_note = key;
			_noteLP = midi_key * 5; //used only during the low-pass filter action
			_midiNote = midi_key;
			if (_note < 0){
				_noteFactor = Math.pow(2,(Math.abs(_note)/12)); //starting from A440 stretch samples this amount.
			} else {
				_noteFactor = Math.pow(0.5,_note/12); //starting from A440 shrink up samples this amount.		
			}
			totalSamples = extractor.totalSamples;
			newLength = Math.floor( totalSamples * _noteFactor );
			noteSamples = new Vector.<Point>(); //new Vector.<Point>(newLength, true); //do not dimention the whole set of samples.
			startFading = (87 - _midiNote) * 9000;
			
			_lastSample = new Point(0, 0);
			_output = new Point(0,0);
			
			_velocity = 127;
			velFact = 1;
			released = true; //is not pressed
			fadeOut = 1;
			fadeIn = 0;
			sindex = 0;
		}
		
		public function get noteIndex():int { return _note; }
		public function get midiNote():int { return _midiNote; }
		public function get noteFactor():int { return _noteFactor; }		
		public function get lastSample():Point { return _lastSample; }
		
		public function pressKey(velocity:uint):void {
			_velocity = velocity;
			velFact = _velocity / 127;
			
			fadeIn = 0;	//fade in the fist 10 samples to avoid
			sindex = 0; //restart the samples head
			
			_playing = true;
			released = false;
			//fadeOut = 1;			
		}
		
		public function reset():void {
			_lastSample.x = _lastSample.y = _output.x = _output.y = 0;
			_playing = false;			 
			_velocity = 127;
			velFact = 1;
			released = true; //is not pressed
			fadeOut = 1;
			fadeIn = 0;
			sindex = 0;
		}
		
		public function releaseKey():void {
			released = true;
			//fadeOut = 1;
		}
		public function get isPressed():Boolean {
			return !released;
		}
		
		/**
		 * It is true whenever the sampler key is generating audio, normally it generates some samples more even after you call the releaseKey method. False when it is completely in silence, note getSample() also may return null.
		 */
		public function get playing():Boolean { return _playing; }
		
		public function getSample():Point {						
			if (sindex < newLength && _playing) {		
				//Alex Nino - yoambulante.com (I've spent some hour of my life in the next lines)
				if (released) { //check if key was just released
					fadeOut =  Math.max(fadeOut - 0.0001, 0);
				} else {
					fadeIn = Math.min(fadeIn + 0.01, 1);
					fadeOut =  fadeIn;
				}
				fades = fadeOut * fadeIn * velFact;
				if (noteSamples.length > sindex && noteSamples[sindex] != null) {					
					_lastSample = noteSamples[sindex].clone();
					_output.x = _lastSample.x * fades;
					_output.y = _lastSample.y * fades;
					sindex++;
					return _output;
				} else {					
					//the key doesn't exists yet, so generate it
					//stretch waveform depending on note pitch
					pos = sindex / newLength * totalSamples;
					i = Math.floor(pos); 					
					step = pos - i; //between 0.0 and 1.0					
					//get samples from extractor.
					s1 = extractor.readSampleAt(i-1);
					s2 = extractor.readSampleAt(i);
					s3 = extractor.readSampleAt(i+1);
					s4 = extractor.readSampleAt(i + 2);		
					//interpolate right channel
					_lastSample.x = HermiteInterpolate(s1.x, s2.x, s3.x, s4.x, step);
					//_lastSample.x = lowpass( _lastSample.x,  hermite,  _midiNote, _noteLP); //lowpass is only use on big set of keys
					//interpolate left channel								
					_lastSample.y = HermiteInterpolate(s1.y, s2.y, s3.y, s4.y, step);
					//_lastSample.y = lowpass( _lastSample.y,  hermite,  _midiNote, _noteLP); //lowpass is only use on big set of keys	
					
					noteSamples[sindex] = _lastSample.clone(); //store the sample value in full volume to be reused after on (earn performance for the next time)	
					//apply fades IN/OUT
					_output.x = _lastSample.x * fades;
					_output.y = _lastSample.y * fades;
					sindex++;
					if (fadeOut > 0) {						
						return _output;
					} else {
						_lastSample.x = _lastSample.y = 0;
						_playing = false;
						return null;
					}					
				}				
			}
			_playing = false;
			return null;
		}
		
		private static function HermiteInterpolate(y0:Number,y1:Number,y2:Number,y3:Number,mu:Number,tension:Number = 0,bias:Number = 0):Number {
			/* Tension: 1 is high, 0 normal, -1 is low
		       Bias: 0 is even, positive is towards first segment, negative towards the other */
			var mu2:Number = mu * mu;
			var mu3:Number = mu2 * mu;
			var m0:Number  = (y1-y0)*(1+bias)*(1-tension)/2;
			m0 += (y2-y1)*(1-bias)*(1-tension)/2;
			var m1:Number  = (y2-y1)*(1+bias)*(1-tension)/2;
			m1 += (y3-y2)*(1-bias)*(1-tension)/2;
			var a0:Number =  2*mu3 - 3*mu2 + 1;
			var a1:Number =    mu3 - 2*mu2 + mu;
			var a2:Number =    mu3 -   mu2;
			var a3:Number = -2*mu3 + 3*mu2;
		
			return (a0*y1+a1*m0+a2*m1+a3*y2);
		}
		
		private static function lowpass(prev:Number, value:Number, dt:Number, rc:Number):Number { // Return RC low-pass filter			
			//see Algorithmic implementation in
			//http://en.wikipedia.org/wiki/Low-pass_filter 
			var alpha:Number = dt / (rc + dt);
			return alpha * value + ( 1 - alpha ) * prev;
		}
	}

}
