package com.yoambulante.midifiles {
	import flash.display.Sprite;
	import flash.events.Event;	
	import flash.events.ProgressEvent;
	import flash.geom.Point;
	import flash.media.Sound;
	import flash.utils.ByteArray;
	
	/**
	* Dispatched when the extraction process has finished. Call <code>.extract</code> method for start extracting all samples.
	*
	* @eventType flash.events.Event.COMPLETE
	*/
	[Event(name = "COMPLETE", type = "flash.events.Event")]
	
	/**
	* Dispatched during the extraction process, helps to know the percentage progress of the current extraction process.
	*
	* @eventType flash.events.ProgressEvent.PROGRESS
	*/
	[Event(name="PROGRESS", type="flash.events.ProgressEvent")]
	
	/**
	 * This class extracts audio samples from a (<code>flash.media.Sound</code>) object and turn them to a Vector.&lt;Point&gt; where .x represents the right channel value and .y the left channel value.
	 * @author Alex Nino
	 */
	public class SamplesExtractor extends Sprite {
		private var samples:Vector.<Point>;
		private var samplesIndex:uint;
		private var _totalSamples:uint;
		/**
		 * Everytime <code>.extract()</code> is called it will try to extract up to this amount of samples
		 */
		public static var MAX_EXTRACTION_PER_GO:uint = 20000; //20.000 by default
		/**
		* @private
		* EMPTY_POINT is for internal use only in same package. 
		*/
		internal static const EMPTY_POINT:Point = new Point(0, 0);		
		
		private var sound:Sound;
		/**
		 * Creates a SamplesExtractor instance
		 * @param	snd <code>Sound</code> object where to extract samples from. IMPORTANT, this sound must represent a waveform in A440 frequency
		 */
		public function SamplesExtractor(snd:Sound) {
			_totalSamples = 0;
			samplesIndex = 0;
			if (snd) {
				sound = snd; 
				_totalSamples = Math.floor(sound.length * 44.1);
				samples = new Vector.<Point>(_totalSamples, true);				
			}
		}
		
		/**
		 * start extracting all samples from a <code>Sound</code> defined on constructor
		 * @author Alex Nino
		 * @see http://www.yoambulante.com/en/downloads/?getfile=playing_pianito
		 * @example The following code extracts samples from a <code>Sound</code>
		 * <listing version="3.0" >
		 * var snd:Sound = new MySound(); //MySound is a sound in the library extending <font>flash.media.Sound</font> Class
		 * var ext:SamplesExtractor = new SamplesExtractor(snd);
		 * ext.addEventListener(ProgressEvent.PROGRESS, extractingSamples); //progress
		 * ext.addEventListener(Event.COMPLETE, samplesReady);	//when is finishes
		 * ext.extract();
		 * </listing>		 
		 */
		public function extract():void {
			this.addEventListener(Event.ENTER_FRAME, extracting);
		}
		
		private function extracting(e:Event):void {
			//try to extract MAX_EXTRACTION_PER_GO each time.
			if (samplesIndex < _totalSamples){
				var total:uint = Math.min(MAX_EXTRACTION_PER_GO, _totalSamples - samplesIndex);	
				var b:ByteArray = new ByteArray();	
				sound.extract(b, total);		
				b.position = 0;
				for (var i:uint = 0; i<total; i++){
					samples[samplesIndex + i] = new Point(b.readFloat(), b.readFloat());
				}
				b.clear();
				b = null;
				samplesIndex += total;
				dispatchEvent(new ProgressEvent(ProgressEvent.PROGRESS, false, false, samplesIndex, _totalSamples));								
			} else {
				//done
				finishExtract();
			}
		}
		
		private function finishExtract():void{
			this.removeEventListener(Event.ENTER_FRAME, extracting);
			sound = null;
			samplesIndex = 0;
			dispatchEvent(new Event(Event.COMPLETE));	
		}
		
		/**
		 * get a stereo sample in form of a <code>Point</code> from the extracted sample list
		 * @param	index
		 * @return a stereo sample in form of a <code>Point</code> where .x represent right channel and .y left channel
		 */
		public function readSampleAt(index:uint):Point {
			if (index >= 0 && index < _totalSamples){				
				return samples[index];
			}
			return EMPTY_POINT; //x:0,y:0
		}
		
		/**
		 * total amount of samples available after extraction
		 */
		public function get totalSamples():uint { return _totalSamples; }
		
	}

}